/*
 * gMapMaker
 * Copyright (C) 2007  Damien Debin
 * http://debin.net/gMapMaker/
 * 
 * This file is part of gMapMaker.
 * 
 * Foobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.IO;
using System.Text.RegularExpressions;
using System.Diagnostics;
using System.ComponentModel;
using System.Threading;
using FreeImageAPI;

namespace gMapMaker
{
  class TileDownloader
  {
    const string MGMapsCacheFolderName = "MGMapsCache";
    const string MGMapsConfFilename = "cache.conf";
    const string MGMapsConfFileContent = "version=3\ntiles_per_file=1\nhash_size={0}\ncenter={1},{2},{3},{4}";

    static string lastProxyUsed = "";
    static int currentProxyRepetition;
    static List<string> proxy_list = new List<string>();

    readonly Random rand = new Random();
    readonly ReportProgressFunction reportFunction;
    readonly BackgroundWorker worker;

    public TileDownloader(ReportProgressFunction reportFunction, BackgroundWorker worker)
    {
      this.reportFunction = reportFunction;
      this.worker = worker;
    }

    void UpdateProxyList()
    {
      reportFunction("Downloading proxy list", 0);

      proxy_list.AddRange(Properties.Settings.Default.Proxys.Split(new char[] { '|' }, StringSplitOptions.RemoveEmptyEntries));

      proxy_list.AddRange(GetProxListFromURLs(Properties.Settings.Default.ProxyListURLs, worker));

      Properties.Settings.Default.Proxys = String.Join("|", proxy_list.ToArray());
    }

    public static List<string> GetProxListFromURLs(string SingleStringProxyListURLs, BackgroundWorker worker)
    {
      List<string> proxylist = new List<string>();
      string[] proxyListURLs = SingleStringProxyListURLs.Split(new char[] { '|' }, StringSplitOptions.RemoveEmptyEntries);

      if (proxyListURLs.Length == 0)
      {
        throw new ArgumentException(Properties.Resources.ExceptionNoProxyList);
      }

      using (WebClient wc = new WebClient())
      {
        foreach (string proxyListURL in proxyListURLs)
        {
          if ((worker != null) && worker.CancellationPending)
          {
            return proxylist;
          }

          if (!proxyListURL.StartsWith("http"))
          {
            proxylist.Add(proxyListURL);
            continue;
          }

          Uri proxyURI = new Uri(proxyListURL);

          wc.Headers[HttpRequestHeader.Referer] = "http://" + proxyURI.Host + "/";
          wc.Headers[HttpRequestHeader.UserAgent] = "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1)";

          int repeatOnError = 0;
          string webPage = null;
          while (String.IsNullOrEmpty(webPage))
          {
            try
            {
              webPage = wc.DownloadString(proxyURI);
            }
            catch (WebException)
            {
              repeatOnError++;
              if (repeatOnError == 3)
              {
                throw;
              }
              Thread.Sleep(500);
              continue;
            }
          }
          
          //webPage = Regex.Replace(webPage, Properties.Settings.Default.ProxyListRegexp_HTMLtoken, string.Empty);

          MatchCollection mc = Regex.Matches(webPage, Properties.Settings.Default.ProxyListRegexp, RegexOptions.None);

          if (mc.Count == 0)
          {
            throw new InvalidOperationException(String.Format(Properties.Resources.ExceptionNoProxyUrlFound, proxyListURL));
          }

          foreach (Match m in mc)
          {
            proxylist.Add(m.Groups["ip"].Value + ':' + m.Groups["port"].Value);
          }
        }
      }

      return proxylist;
    }

    void DownloadImage(string url, string filepath, bool useProxy)
    {
      bool error_downloading;
      int download_try = 0;

      if (File.Exists(filepath))
      {
        FileInfo fi = new FileInfo(filepath);
        if (fi.Length > 0)
        {
          return;
        }
        File.Delete(filepath);
      }

      if (useProxy && (proxy_list.Count == 0))
      {
        UpdateProxyList();
      }

      using (WebClient wc = new WebClient())
      {
        while (true)
        {
          if (worker.CancellationPending) return;

          wc.Headers[HttpRequestHeader.UserAgent] = "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; .NET CLR 1.0.3705;)";
          wc.Headers[HttpRequestHeader.Referer] = Properties.Settings.Default.GoogleReferrer;

          if (useProxy && !String.IsNullOrEmpty(Properties.Settings.Default.ProxyListURLs))
          {
            if (proxy_list.Count == 0)
            {
              if (Properties.Settings.Default.HaltOnProxyListExhausted)
              {
                throw new InvalidOperationException(Properties.Resources.ExceptionProxyListExhausted);
              }
              else
              {
                UpdateProxyList();
              }
            }
            if (String.IsNullOrEmpty(lastProxyUsed))
            {
              lastProxyUsed = proxy_list[rand.Next(proxy_list.Count)];
              currentProxyRepetition = 0;
            }
            wc.Proxy = new WebProxy(lastProxyUsed);
            currentProxyRepetition++;
          }

          error_downloading = false;
          try
          {
            download_try++;
            reportFunction("Downloading " + Path.GetFileName(filepath) + " (try " + download_try + ") " + (useProxy ? " from proxy " + lastProxyUsed : ""), 0);
            wc.DownloadFile(url, filepath);
          }
          catch (WebException e)
          {
            Debug.WriteLine(String.Format("Proxy {0} ; error: {1} ; message: {2}", lastProxyUsed, e.Message, (e.Response != null ? ((HttpWebResponse)e.Response).StatusCode.ToString() : "n/a")));
            if ((e.Response != null) && (((HttpWebResponse)e.Response).StatusCode == HttpStatusCode.NotFound))
            {
              //image does not exist in GG database
              break;
            }
            else
            {
              error_downloading = true;
            }
          }

          if (error_downloading || ((wc.ResponseHeaders[HttpResponseHeader.Server] != null) && wc.ResponseHeaders[HttpResponseHeader.Server].StartsWith("GCS")))
          {
            if (!String.IsNullOrEmpty(lastProxyUsed))
            {
              proxy_list.Remove(lastProxyUsed);
              Properties.Settings.Default.Proxys = String.Join("|", proxy_list.ToArray());
              lastProxyUsed = "";
              continue;
            }
            else
            {
              throw new InvalidOperationException(Properties.Resources.ExceptionDownloadingImage);
            }
          }

          if (currentProxyRepetition >= 5)
          {
            //change every 3 requests to the same proxy, avoid proxy overload
            lastProxyUsed = "";
          }

          if (Properties.Settings.Default.SlowDown)
          {
            Thread.Sleep((int)Properties.Settings.Default.SlowDownDelay); //in ms
          }

          break;
        }
      }
    }

    public string DownloadTile(string tileType, GMapTile g)
    {
      string src_filename = Properties.Settings.Default.CacheFullPath, url, mgmapsmode = "";
      bool useProxy = false;

      if (g.Zoom > 22) return "";

      if (Properties.Settings.Default.MGMapsMode && !src_filename.EndsWith(MGMapsCacheFolderName))
      {
        src_filename = Path.Combine(src_filename, MGMapsCacheFolderName);
      }

      if (Properties.Settings.Default.MGMapsHashSize < 1)
      {
        throw new InvalidOperationException(Properties.Resources.ExceptionHashSizeNegative);
      }

      switch (tileType)
      {
        default:
        case "kh":
          url = Properties.Settings.Default.GoogleKH;
          url = url.Replace("QQQQ", g.Quadtree);
          if (Properties.Settings.Default.MGMapsMode)
          {
            mgmapsmode = "GoogleSat";
            src_filename = Path.Combine(src_filename, "GoogleSat_" + g.Zoom.ToString());
            src_filename = Path.Combine(src_filename, ((g.X * 256 + g.Y) % Properties.Settings.Default.MGMapsHashSize).ToString());
            src_filename = Path.Combine(src_filename, String.Format("{0:d}_{1:d}.mgm", g.X, g.Y));
          }
          else
          {
            src_filename = Path.Combine(src_filename, "kh");
            src_filename = Path.Combine(src_filename, (g.Zoom + 1).ToString());
            src_filename = Path.Combine(src_filename, g.Quadtree + ".jpg");
          }
          useProxy = Properties.Settings.Default.UseProxyForSatelliteDownload;
          break;
        case "mt":
          url = Properties.Settings.Default.GoogleMT;
          url = url.Replace("XXXX", g.X.ToString());
          url = url.Replace("YYYY", g.Y.ToString());
          url = url.Replace("ZZZZ", (17 - (int)g.Zoom).ToString());
          if (Properties.Settings.Default.MGMapsMode)
          {
            mgmapsmode = "GoogleMap";
            src_filename = Path.Combine(src_filename, "GoogleMap_" + g.Zoom.ToString());
            src_filename = Path.Combine(src_filename, ((g.X * 256 + g.Y) % Properties.Settings.Default.MGMapsHashSize).ToString());
            src_filename = Path.Combine(src_filename, String.Format("{0:d}_{1:d}.mgm", g.X, g.Y));
          }
          else
          {
            src_filename = Path.Combine(src_filename, "mt");
            src_filename = Path.Combine(src_filename, (g.Zoom + 1).ToString());
            src_filename = Path.Combine(src_filename, g.Quadtree + ".png");
          }
          useProxy = Properties.Settings.Default.UseProxyForDrivingMapDownload;
          break;
        case "tt":
          url = Properties.Settings.Default.GoogleTT;
          url = url.Replace("XXXX", g.X.ToString());
          url = url.Replace("YYYY", g.Y.ToString());
          url = url.Replace("ZZZZ", (17 - (int)g.Zoom).ToString());
          if (Properties.Settings.Default.MGMapsMode)
          {
            mgmapsmode = "GoogleHyb";
            src_filename = Path.Combine(src_filename, "GoogleHyb_" + g.Zoom.ToString());
            src_filename = Path.Combine(src_filename, ((g.X * 256 + g.Y) % Properties.Settings.Default.MGMapsHashSize).ToString());
            src_filename = Path.Combine(src_filename, String.Format("{0:d}_{1:d}.mgm", g.X, g.Y));
          }
          else
          {
            src_filename = Path.Combine(src_filename, "tt");
            src_filename = Path.Combine(src_filename, (g.Zoom + 1).ToString());
            src_filename = Path.Combine(src_filename, g.Quadtree + ".png");
          }
          useProxy = Properties.Settings.Default.UseProxyForDrivingMapDownload;
          break;
        case "pt":
          url = Properties.Settings.Default.GooglePT;
          url = url.Replace("XXXX", g.X.ToString());
          url = url.Replace("YYYY", g.Y.ToString());
          url = url.Replace("ZZZZ", (17 - (int)g.Zoom).ToString());
          if (Properties.Settings.Default.MGMapsMode)
          {
            mgmapsmode = "GoogleTer";
            src_filename = Path.Combine(src_filename, "GoogleTer_" + g.Zoom.ToString());
            src_filename = Path.Combine(src_filename, ((g.X * 256 + g.Y) % Properties.Settings.Default.MGMapsHashSize).ToString());
            src_filename = Path.Combine(src_filename, String.Format("{0:d}_{1:d}.mgm", g.X, g.Y));
          }
          else
          {
            src_filename = Path.Combine(src_filename, "pt");
            src_filename = Path.Combine(src_filename, (g.Zoom + 1).ToString());
            src_filename = Path.Combine(src_filename, g.Quadtree + ".png");
          }
          useProxy = Properties.Settings.Default.UseProxyForDrivingMapDownload;
          break;
      }
      url = url.Replace("*", rand.Next(4).ToString());

      string downloadedFilePath = Path.GetDirectoryName(src_filename);
      if (!Directory.Exists(downloadedFilePath))
      {
        Directory.CreateDirectory(downloadedFilePath);
      }

      //check if conf file exists in (MGMaps) cache folder
      string MGMapsConfFileFullPath = Path.Combine(downloadedFilePath, "../../" + MGMapsConfFilename);
      if (Properties.Settings.Default.MGMapsMode && !File.Exists(MGMapsConfFileFullPath))
      {
        //creates MGMaps cache .conf file
        double lat = (g.TopLat - g.BottomLat) / 2;
        double lng = (g.RightLong - g.LeftLong) / 2;
        File.WriteAllText(MGMapsConfFileFullPath, String.Format(MGMapsConfFileContent, Properties.Settings.Default.MGMapsHashSize, lat.ToString(MainForm.ciUS.NumberFormat), lng.ToString(MainForm.ciUS.NumberFormat), g.Zoom, mgmapsmode), Encoding.ASCII);
      }

      DownloadImage(url, src_filename, useProxy);

      return src_filename;
    }
  }
}
